@@ -17,10 +17,9 @@
  */
 
 #include "config.h"
-#include <math.h>
+#include <cmath>
 
 #ifndef __GI_SCANNER__
-#include <goo/GooList.h>
 #include <GlobalParams.h>
 #include <PDFDoc.h>
 #include <Outline.h>
@@ -28,11 +27,14 @@
 #include <UnicodeMap.h>
 #include <GfxState.h>
 #include <PageTransition.h>
+#include <BBoxOutputDev.h>
 #endif
 
 #include "poppler.h"
 #include "poppler-private.h"
 
+#define SUPPORTED_ROTATION(r) (r == 90 || r == 180 || r == 270)
+
 /**
  * SECTION:poppler-page
  * @short_description: Information about a page in a document
@@ -60,7 +62,7 @@ _poppler_page_new (PopplerDocument *document, Page *page, int index)
 
   g_return_val_if_fail (POPPLER_IS_DOCUMENT (document), NULL);
 
-  poppler_page = (PopplerPage *) g_object_new (POPPLER_TYPE_PAGE, NULL, NULL);
+  poppler_page = (PopplerPage *) g_object_new (POPPLER_TYPE_PAGE, nullptr, NULL);
   poppler_page->document = (PopplerDocument *) g_object_ref (document);
   poppler_page->page = page;
   poppler_page->index = index;
@@ -74,9 +76,9 @@ poppler_page_finalize (GObject *object)
   PopplerPage *page = POPPLER_PAGE (object);
 
   g_object_unref (page->document);
-  page->document = NULL;
+  page->document = nullptr;
 
-  if (page->text != NULL) 
+  if (page->text != nullptr) 
     page->text->decRefCnt();
   /* page->page is owned by the document */
 
@@ -110,9 +112,9 @@ poppler_page_get_size (PopplerPage *page,
     page_height = page->page->getCropHeight ();
   }
 
-  if (width != NULL)
+  if (width != nullptr)
     *width = page_width;
-  if (height != NULL)
+  if (height != nullptr)
     *height = page_height;
 }
 
@@ -184,16 +186,15 @@ poppler_page_get_transition (PopplerPage *page)
 {
   PageTransition *trans;
   PopplerPageTransition *transition;
-  Object obj;
-  
+
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
 
-  trans = new PageTransition (page->page->getTrans (&obj));
-  obj.free ();
+  Object obj = page->page->getTrans ();
+  trans = new PageTransition (&obj);
 
   if (!trans->isOk ()) {
     delete trans;
-    return NULL;
+    return nullptr;
   }
 
   transition = poppler_page_transition_new ();
@@ -249,6 +250,7 @@ poppler_page_get_transition (PopplerPage *page)
 	  POPPLER_PAGE_TRANSITION_OUTWARD;
 
   transition->duration = trans->getDuration();
+  transition->duration_real = trans->getDuration();
   transition->angle = trans->getAngle();
   transition->scale = trans->getScale();
   transition->rectangular = trans->isRectangular();
@@ -261,18 +263,18 @@ poppler_page_get_transition (PopplerPage *page)
 static TextPage *
 poppler_page_get_text_page (PopplerPage *page)
 {
-  if (page->text == NULL) {
+  if (page->text == nullptr) {
     TextOutputDev *text_dev;
     Gfx           *gfx;
 
-    text_dev = new TextOutputDev (NULL, gTrue, 0, gFalse, gFalse);
+    text_dev = new TextOutputDev (nullptr, true, 0, false, false);
     gfx = page->page->createGfx(text_dev,
 				72.0, 72.0, 0,
-				gFalse, /* useMediaBox */
-				gTrue, /* Crop */
+				false, /* useMediaBox */
+				true, /* Crop */
 				-1, -1, -1, -1,
-				gFalse, /* printing */
-				NULL, NULL);
+				false, /* printing */
+				nullptr, nullptr);
     page->page->display(gfx);
     text_dev->endPage();
 
@@ -304,13 +306,13 @@ annot_is_markup (Annot *annot)
     }
 }
 
-static GBool
+static bool
 poppler_print_annot_cb (Annot *annot, void *user_data)
 {
   PopplerPrintFlags user_print_flags = (PopplerPrintFlags)GPOINTER_TO_INT (user_data);
 
   if (annot->getFlags () & Annot::flagHidden)
-    return gFalse;
+    return false;
 
   if (user_print_flags & POPPLER_PRINT_STAMP_ANNOTS_ONLY) {
     return (annot->getType() == Annot::typeStamp) ?
@@ -331,7 +333,7 @@ poppler_print_annot_cb (Annot *annot, void *user_data)
 static void
 _poppler_page_render (PopplerPage      *page,
 		      cairo_t          *cairo,
-		      GBool             printing,
+		      bool             printing,
                       PopplerPrintFlags print_flags)
 {
   CairoOutputDev *output_dev;
@@ -343,8 +345,8 @@ _poppler_page_render (PopplerPage      *page,
   output_dev->setPrinting (printing);
 
 
-  if (!printing && page->text == NULL) {
-    page->text = new TextPage (gFalse);
+  if (!printing && page->text == nullptr) {
+    page->text = new TextPage (false);
     output_dev->setTextPage (page->text);
   }
   /* NOTE: instead of passing -1 we should/could use cairo_clip_extents()
@@ -352,18 +354,18 @@ _poppler_page_render (PopplerPage      *page,
   cairo_save (cairo);
   page->page->displaySlice(output_dev,
 			   72.0, 72.0, 0,
-			   gFalse, /* useMediaBox */
-			   gTrue, /* Crop */
+			   false, /* useMediaBox */
+			   true, /* Crop */
 			   -1, -1,
 			   -1, -1,
 			   printing,
-			   NULL, NULL,
-			   printing ? poppler_print_annot_cb : NULL,
-                           printing ? GINT_TO_POINTER ((gint)print_flags) : NULL);
+			   nullptr, nullptr,
+			   printing ? poppler_print_annot_cb : nullptr,
+                           printing ? GINT_TO_POINTER ((gint)print_flags) : nullptr);
   cairo_restore (cairo);
 
-  output_dev->setCairo (NULL);
-  output_dev->setTextPage (NULL);
+  output_dev->setCairo (nullptr);
+  output_dev->setTextPage (nullptr);
 }
 
 /**
@@ -374,7 +376,9 @@ _poppler_page_render (PopplerPage      *page,
  * Render the page to the given cairo context. This function
  * is for rendering a page that will be displayed. If you want
  * to render a page that will be printed use
- * poppler_page_render_for_printing() instead
+ * poppler_page_render_for_printing() instead.  Please see the documentation
+ * for that function for the differences between rendering to the screen and
+ * rendering to a printer.
  **/
 void
 poppler_page_render (PopplerPage *page,
@@ -382,7 +386,7 @@ poppler_page_render (PopplerPage *page,
 {
   g_return_if_fail (POPPLER_IS_PAGE (page));
 
-  _poppler_page_render (page, cairo, gFalse, (PopplerPrintFlags)0);
+  _poppler_page_render (page, cairo, false, (PopplerPrintFlags)0);
 }
 
 /**
@@ -394,6 +398,9 @@ poppler_page_render (PopplerPage *page,
  * Render the page to the given cairo context for printing
  * with the specified options
  *
+ * See the documentation for poppler_page_render_for_printing() for the
+ * differences between rendering to the screen and rendering to a printer.
+ *
  * Since: 0.16
  **/
 void
@@ -403,7 +410,7 @@ poppler_page_render_for_printing_with_options (PopplerPage      *page,
 {
   g_return_if_fail (POPPLER_IS_PAGE (page));
 
-  _poppler_page_render (page, cairo, gTrue, options);
+  _poppler_page_render (page, cairo, true, options);
 }
 
 /**
@@ -411,7 +418,36 @@ poppler_page_render_for_printing_with_options (PopplerPage      *page,
  * @page: the page to render from
  * @cairo: cairo context to render to
  *
- * Render the page to the given cairo context for printing.
+ * Render the page to the given cairo context for printing with
+ * #POPPLER_PRINT_ALL flags selected.  If you want a different set of flags,
+ * use poppler_page_render_for_printing_with_options().
+ *
+ * The difference between poppler_page_render() and this function is that some
+ * things get rendered differently between screens and printers:
+ *
+ * <itemizedlist>
+ *   <listitem>
+ *     PDF annotations get rendered according to their #PopplerAnnotFlag value.
+ *     For example, #POPPLER_ANNOT_FLAG_PRINT refers to whether an annotation
+ *     is printed or not, whereas #POPPLER_ANNOT_FLAG_NO_VIEW refers to whether
+ *     an annotation is invisible when displaying to the screen.
+ *   </listitem>
+ *   <listitem>
+ *     PDF supports "hairlines" of width 0.0, which often get rendered as
+ *     having a width of 1 device pixel.  When displaying on a screen, Cairo
+ *     may render such lines wide so that they are hard to see, and Poppler
+ *     makes use of PDF's Stroke Adjust graphics parameter to make the lines
+ *     easier to see.  However, when printing, Poppler is able to directly use a
+ *     printer's pixel size instead.
+ *   </listitem>
+ *   <listitem>
+ *     Some advanced features in PDF may require an image to be rasterized
+ *     before sending off to a printer.  This may produce raster images which
+ *     exceed Cairo's limits.  The "printing" functions will detect this condition
+ *     and try to down-scale the intermediate surfaces as appropriate.
+ *   </listitem>
+ * </itemizedlist>
+ * 
  **/
 void
 poppler_page_render_for_printing (PopplerPage *page,
@@ -419,7 +455,7 @@ poppler_page_render_for_printing (PopplerPage *page,
 {
   g_return_if_fail (POPPLER_IS_PAGE (page));
 
-  _poppler_page_render (page, cairo, gTrue, POPPLER_PRINT_ALL);
+  _poppler_page_render (page, cairo, true, POPPLER_PRINT_ALL);
 }
 
 static cairo_surface_t *
@@ -435,7 +471,7 @@ create_surface_from_thumbnail_data (guchar *data,
 
   surface = cairo_image_surface_create (CAIRO_FORMAT_RGB24, width, height);
   if (cairo_surface_status (surface))
-    return NULL;
+    return nullptr;
 
   cairo_pixels = cairo_image_surface_get_data (surface);
   cairo_stride = cairo_image_surface_get_stride (surface);
@@ -488,7 +524,7 @@ poppler_page_get_thumbnail (PopplerPage *page)
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
 
   if (!page->page->loadThumb (&data, &width, &height, &rowstride))
-    return NULL;
+    return nullptr;
 
   surface = create_surface_from_thumbnail_data (data, width, height, rowstride);
   gfree (data);
@@ -566,7 +602,7 @@ poppler_page_render_selection (PopplerPage           *page,
 		       &pdf_selection, selection_style,
 		       &gfx_glyph_color, &gfx_background_color);
 
-  output_dev->setCairo (NULL);
+  output_dev->setCairo (nullptr);
 }
 
 /**
@@ -587,18 +623,16 @@ poppler_page_get_thumbnail_size (PopplerPage *page,
 				 int         *width,
 				 int         *height)
 {
-  Object thumb;
   Dict *dict;
   gboolean retval = FALSE;
 
   g_return_val_if_fail (POPPLER_IS_PAGE (page), FALSE);
-  g_return_val_if_fail (width != NULL, FALSE);
-  g_return_val_if_fail (height != NULL, FALSE);
+  g_return_val_if_fail (width != nullptr, FALSE);
+  g_return_val_if_fail (height != nullptr, FALSE);
 
-  page->page->getThumb (&thumb);
+  Object thumb = page->page->getThumb ();
   if (!thumb.isStream ())
     {
-      thumb.free ();
       return FALSE;
     }
 
@@ -610,8 +644,6 @@ poppler_page_get_thumbnail_size (PopplerPage *page,
       dict->lookupInt ("Height", "H", height))
     retval = TRUE;
 
-  thumb.free ();
-
   return retval;
 }
 
@@ -640,9 +672,7 @@ poppler_page_get_selection_region (PopplerPage           *page,
   PDFRectangle poppler_selection;
   TextPage *text;
   SelectionStyle selection_style = selectionStyleGlyph;
-  GooList *list;
-  GList *region = NULL;
-  int i;
+  GList *region = nullptr;
 
   poppler_selection.x1 = selection->x1;
   poppler_selection.y1 = selection->y1;
@@ -663,11 +693,10 @@ poppler_page_get_selection_region (PopplerPage           *page,
     }
 
   text = poppler_page_get_text_page (page);
-  list = text->getSelectionRegion(&poppler_selection,
+  std::vector<PDFRectangle*>* list = text->getSelectionRegion(&poppler_selection,
 				  selection_style, scale);
 
-  for (i = 0; i < list->getLength(); i++) {
-    PDFRectangle *selection_rect = (PDFRectangle *) list->get(i);
+  for (const PDFRectangle *selection_rect : *list) {
     PopplerRectangle *rect;
 
     rect = poppler_rectangle_new ();
@@ -702,8 +731,7 @@ poppler_page_selection_region_free (GList *region)
   if (G_UNLIKELY (!region))
     return;
 
-  g_list_foreach (region, (GFunc)poppler_rectangle_free, NULL);
-  g_list_free (region);
+  g_list_free_full (region, (GDestroyNotify)poppler_rectangle_free);
 }
 
 /**
@@ -730,9 +758,7 @@ poppler_page_get_selected_region (PopplerPage           *page,
   PDFRectangle poppler_selection;
   TextPage *text;
   SelectionStyle selection_style = selectionStyleGlyph;
-  GooList *list;
   cairo_region_t *region;
-  int i;
 
   poppler_selection.x1 = selection->x1;
   poppler_selection.y1 = selection->y1;
@@ -753,13 +779,12 @@ poppler_page_get_selected_region (PopplerPage           *page,
     }
 
   text = poppler_page_get_text_page (page);
-  list = text->getSelectionRegion(&poppler_selection,
+  std::vector<PDFRectangle*>* list = text->getSelectionRegion(&poppler_selection,
 				  selection_style, 1.0);
 
   region = cairo_region_create ();
 
-  for (i = 0; i < list->getLength(); i++) {
-    PDFRectangle *selection_rect = (PDFRectangle *) list->get(i);
+  for (const PDFRectangle *selection_rect : *list) {
     cairo_rectangle_int_t rect;
 
     rect.x = (gint) ((selection_rect->x1 * scale) + 0.5);
@@ -800,7 +825,7 @@ poppler_page_get_selected_text (PopplerPage          *page,
   PDFRectangle pdf_selection;
 
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
-  g_return_val_if_fail (selection != NULL, NULL);
+  g_return_val_if_fail (selection != nullptr, NULL);
 
   pdf_selection.x1 = selection->x1;
   pdf_selection.y1 = selection->y1;
@@ -822,7 +847,7 @@ poppler_page_get_selected_text (PopplerPage          *page,
 
   text = poppler_page_get_text_page (page);
   sel_text = text->getSelectionText (&pdf_selection, selection_style);
-  result = g_strdup (sel_text->getCString ());
+  result = g_strdup (sel_text->c_str ());
   delete sel_text;
 
   return result;
@@ -866,7 +891,7 @@ poppler_page_get_text_for_area (PopplerPage      *page,
                                 PopplerRectangle *area)
 {
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
-  g_return_val_if_fail (area != NULL, NULL);
+  g_return_val_if_fail (area != nullptr, NULL);
 
   return poppler_page_get_selected_text (page, POPPLER_SELECTION_GLYPH, area);
 }
@@ -879,7 +904,7 @@ poppler_page_get_text_for_area (PopplerPage      *page,
  * @options: find options
  *
  * Finds @text in @page with the given #PopplerFindFlags options and
- * returns a #GList of rectangles for each occurance of the text on the page.
+ * returns a #GList of rectangles for each occurrence of the text on the page.
  * The coordinates are in PDF points.
  *
  * Return value: (element-type PopplerRectangle) (transfer full): a #GList of #PopplerRectangle,
@@ -902,23 +927,24 @@ poppler_page_find_text_with_options (PopplerPage     *page,
   gboolean start_at_last = FALSE;
 
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
-  g_return_val_if_fail (text != NULL, NULL);
+  g_return_val_if_fail (text != nullptr, NULL);
 
   text_dev = poppler_page_get_text_page (page);
 
   ucs4 = g_utf8_to_ucs4_fast (text, -1, &ucs4_len);
-  poppler_page_get_size (page, NULL, &height);
+  poppler_page_get_size (page, nullptr, &height);
 
   backwards = options & POPPLER_FIND_BACKWARDS;
-  matches = NULL;
+  matches = nullptr;
   xMin = 0;
   yMin = backwards ? height : 0;
 
   while (text_dev->findText (ucs4, ucs4_len,
-                             gFalse, gTrue, // startAtTop, stopAtBottom
+                             false, true, // startAtTop, stopAtBottom
                              start_at_last,
-                             gFalse, //stopAtLast
+                             false, //stopAtLast
                              options & POPPLER_FIND_CASE_SENSITIVE,
+                             options & POPPLER_FIND_IGNORE_DIACRITICS,
                              backwards,
                              options & POPPLER_FIND_WHOLE_WORDS_ONLY,
                              &xMin, &yMin, &xMax, &yMax))
@@ -943,7 +969,7 @@ poppler_page_find_text_with_options (PopplerPage     *page,
  * @text: the text to search for (UTF-8 encoded)
  *
  * Finds @text in @page with the default options (%POPPLER_FIND_DEFAULT) and
- * returns a #GList of rectangles for each occurance of the text on the page.
+ * returns a #GList of rectangles for each occurrence of the text on the page.
  * The coordinates are in PDF points.
  *
  * Return value: (element-type PopplerRectangle) (transfer full): a #GList of #PopplerRectangle,
@@ -957,7 +983,7 @@ poppler_page_find_text (PopplerPage *page,
 
 static CairoImageOutputDev *
 poppler_page_get_image_output_dev (PopplerPage *page,
-				   GBool (*imgDrawDeviceCbk)(int img_id, void *data),
+				   bool (*imgDrawDeviceCbk)(int img_id, void *data),
 				   void *imgDrawCbkData)
 {
   CairoImageOutputDev *image_dev;
@@ -972,11 +998,11 @@ poppler_page_get_image_output_dev (PopplerPage *page,
 
   gfx = page->page->createGfx(image_dev,
 			      72.0, 72.0, 0,
-			      gFalse, /* useMediaBox */
-			      gTrue, /* Crop */
+			      false, /* useMediaBox */
+			      true, /* Crop */
 			      -1, -1, -1, -1,
-			      gFalse, /* printing */
-			      NULL, NULL);
+			      false, /* printing */
+			      nullptr, nullptr);
   page->page->display(gfx);
   delete gfx;
 
@@ -996,13 +1022,13 @@ poppler_page_get_image_output_dev (PopplerPage *page,
 GList *
 poppler_page_get_image_mapping (PopplerPage *page)
 {
-  GList *map_list = NULL;
+  GList *map_list = nullptr;
   CairoImageOutputDev *out;
   gint i;
   
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
 
-  out = poppler_page_get_image_output_dev (page, NULL, NULL);
+  out = poppler_page_get_image_output_dev (page, nullptr, nullptr);
 
   for (i = 0; i < out->getNumImages (); i++) {
     PopplerImageMapping *mapping;
@@ -1030,7 +1056,7 @@ poppler_page_get_image_mapping (PopplerPage *page)
   return map_list;	
 }
 
-static GBool
+static bool
 image_draw_decide_cb (int image_id, void *data)
 {
   return (image_id == GPOINTER_TO_INT (data));
@@ -1061,14 +1087,14 @@ poppler_page_get_image (PopplerPage *page,
   if (image_id >= out->getNumImages ()) {
     delete out;
     
-    return NULL;
+    return nullptr;
   }
 
   image = out->getImage (image_id)->getImage ();
   if (!image) {
     delete out;
 
-    return NULL;
+    return nullptr;
   }
 
   cairo_surface_reference (image);
@@ -1088,11 +1114,10 @@ poppler_page_get_image (PopplerPage *page,
 void
 poppler_page_free_image_mapping (GList *list)
 {
-  if (G_UNLIKELY (list == NULL))
+  if (G_UNLIKELY (list == nullptr))
     return;
 
-  g_list_foreach (list, (GFunc)poppler_image_mapping_free, NULL);
-  g_list_free (list);
+  g_list_free_full (list, (GDestroyNotify)poppler_image_mapping_free);
 }
 
 /**
@@ -1108,7 +1133,7 @@ poppler_page_render_to_ps (PopplerPage   *page,
 			   PopplerPSFile *ps_file)
 {
   g_return_if_fail (POPPLER_IS_PAGE (page));
-  g_return_if_fail (ps_file != NULL);
+  g_return_if_fail (ps_file != nullptr);
 
   if (!ps_file->out)  {
     std::vector<int> pages;
@@ -1117,15 +1142,15 @@ poppler_page_render_to_ps (PopplerPage   *page,
     }
     ps_file->out = new PSOutputDev (ps_file->filename,
                                     ps_file->document->doc,
-                                    NULL, pages,
+                                    nullptr, pages,
                                     psModePS, (int)ps_file->paper_width,
                                     (int)ps_file->paper_height, ps_file->duplex,
-                                    0, 0, 0, 0, gFalse, gFalse);
+                                    false, 0, 0, 0, false, false);
   }
 
 
   ps_file->document->doc->displayPage (ps_file->out, page->index + 1, 72.0, 72.0,
-				       0, gFalse, gTrue, gFalse);
+				       0, false, true, false);
 }
 
 static void
@@ -1164,7 +1189,7 @@ poppler_page_class_init (PopplerPageClass *klass)
 				   g_param_spec_string ("label",
 							"Page Label",
 							"The label of the page",
-							NULL,
+							nullptr,
 							G_PARAM_READABLE));
 }
 
@@ -1186,7 +1211,7 @@ poppler_page_init (PopplerPage *page)
 GList *
 poppler_page_get_link_mapping (PopplerPage *page)
 {
-  GList *map_list = NULL;
+  GList *map_list = nullptr;
   gint i;
   Links *links;
   double width, height;
@@ -1195,8 +1220,8 @@ poppler_page_get_link_mapping (PopplerPage *page)
   
   links = new Links (page->page->getAnnots ());
 
-  if (links == NULL)
-    return NULL;
+  if (links == nullptr)
+    return nullptr;
   
   poppler_page_get_size (page, &width, &height);
   
@@ -1212,7 +1237,7 @@ poppler_page_get_link_mapping (PopplerPage *page)
       
       /* Create the mapping */
       mapping = poppler_link_mapping_new ();
-      mapping->action = _poppler_action_new (page->document, link_action, NULL);
+      mapping->action = _poppler_action_new (page->document, link_action, nullptr);
 
       link->getRect (&rect.x1, &rect.y1, &rect.x2, &rect.y2);
 
@@ -1272,11 +1297,10 @@ poppler_page_get_link_mapping (PopplerPage *page)
 void
 poppler_page_free_link_mapping (GList *list)
 {
-  if (G_UNLIKELY (list == NULL))
+  if (G_UNLIKELY (list == nullptr))
     return;
 
-  g_list_foreach (list, (GFunc)poppler_link_mapping_free, NULL);
-  g_list_free (list);
+  g_list_free_full (list, (GDestroyNotify)poppler_link_mapping_free);
 }
 
 /**
@@ -1292,7 +1316,7 @@ poppler_page_free_link_mapping (GList *list)
 GList *
 poppler_page_get_form_field_mapping (PopplerPage *page)
 {
-  GList *map_list = NULL;
+  GList *map_list = nullptr;
   FormPageWidgets *forms;
   gint i;
   
@@ -1300,8 +1324,8 @@ poppler_page_get_form_field_mapping (PopplerPage *page)
 
   forms = page->page->getFormWidgets ();
 
-  if (forms == NULL)
-    return NULL;
+  if (forms == nullptr)
+    return nullptr;
   
   for (i = 0; i < forms->getNumWidgets (); i++) {
     PopplerFormFieldMapping *mapping;
@@ -1339,11 +1363,10 @@ poppler_page_get_form_field_mapping (PopplerPage *page)
 void
 poppler_page_free_form_field_mapping (GList *list)
 {
-  if (G_UNLIKELY (list == NULL))
+  if (G_UNLIKELY (list == nullptr))
     return;
 
-  g_list_foreach (list, (GFunc) poppler_form_field_mapping_free, NULL);
-  g_list_free (list);
+  g_list_free_full (list, (GDestroyNotify)poppler_form_field_mapping_free);
 }
 
 /**
@@ -1359,27 +1382,31 @@ poppler_page_free_form_field_mapping (GList *list)
 GList *
 poppler_page_get_annot_mapping (PopplerPage *page)
 {
-  GList *map_list = NULL;
+  GList *map_list = nullptr;
   double width, height;
   gint i;
   Annots *annots;
+  const PDFRectangle *crop_box;
 
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
 
   annots = page->page->getAnnots ();
   if (!annots)
-    return NULL;
+    return nullptr;
 
   poppler_page_get_size (page, &width, &height);
+  crop_box = page->page->getCropBox ();
 
   for (i = 0; i < annots->getNumAnnots (); i++) {
     PopplerAnnotMapping *mapping;
     PopplerRectangle rect;
     Annot *annot;
     PDFRectangle *annot_rect;
+    gboolean flag_no_rotate;
     gint rotation = 0;
 
     annot = annots->getAnnot (i);
+    flag_no_rotate = annot->getFlags () & Annot::flagNoRotate;
 
     /* Create the mapping */
     mapping = poppler_annot_mapping_new ();
@@ -1422,40 +1449,58 @@ poppler_page_get_annot_mapping (PopplerPage *page)
       }
 
     annot_rect = annot->getRect ();
-    rect.x1 = annot_rect->x1 - page->page->getCropBox()->x1;
-    rect.y1 = annot_rect->y1 - page->page->getCropBox()->y1;
-    rect.x2 = annot_rect->x2 - page->page->getCropBox()->x1;
-    rect.y2 = annot_rect->y2 - page->page->getCropBox()->y1;
+    rect.x1 = annot_rect->x1 - crop_box->x1;
+    rect.y1 = annot_rect->y1 - crop_box->y1;
+    rect.x2 = annot_rect->x2 - crop_box->x1;
+    rect.y2 = annot_rect->y2 - crop_box->y1;
 
-    if (! (annot->getFlags () & Annot::flagNoRotate))
-      rotation = page->page->getRotate ();
+    rotation = page->page->getRotate ();
 
-    switch (rotation)
-      {
-      case 90:
-        mapping->area.x1 = rect.y1;
-        mapping->area.y1 = height - rect.x2;
-        mapping->area.x2 = mapping->area.x1 + (rect.y2 - rect.y1);
-        mapping->area.y2 = mapping->area.y1 + (rect.x2 - rect.x1);
-        break;
-      case 180:
-        mapping->area.x1 = width - rect.x2;
-        mapping->area.y1 = height - rect.y2;
-        mapping->area.x2 = mapping->area.x1 + (rect.x2 - rect.x1);
-        mapping->area.y2 = mapping->area.y1 + (rect.y2 - rect.y1);
-        break;
-      case 270:
-        mapping->area.x1 = width - rect.y2;
-        mapping->area.y1 = rect.x1;
-        mapping->area.x2 = mapping->area.x1 + (rect.y2 - rect.y1);
-        mapping->area.y2 = mapping->area.y1 + (rect.x2 - rect.x1);
-        break;
-      default:
-        mapping->area.x1 = rect.x1;
-        mapping->area.y1 = rect.y1;
-        mapping->area.x2 = rect.x2;
-        mapping->area.y2 = rect.y2;
+    if (rotation == 0 || !SUPPORTED_ROTATION (rotation)) { /* zero or unknown rotation */
+      mapping->area.x1 = rect.x1;
+      mapping->area.y1 = rect.y1;
+      mapping->area.x2 = rect.x2;
+      mapping->area.y2 = rect.y2;
+    } else {
+      gdouble annot_height = rect.y2 - rect.y1;
+      gdouble annot_width = rect.x2 - rect.x1;
+
+      if (flag_no_rotate) {
+        if (rotation == 90) {
+          mapping->area.x1 = rect.y2;
+          mapping->area.y1 = height - (rect.x1 + annot_height);
+          mapping->area.x2 = rect.y2 + annot_width;
+          mapping->area.y2 = height - rect.x1;
+        } else if (rotation == 180) {
+          mapping->area.x1 = width - rect.x1;
+          mapping->area.x2 = MIN (mapping->area.x1 + annot_width, width);
+          mapping->area.y2 = height - rect.y2;
+          mapping->area.y1 = MAX (0, mapping->area.y2 - annot_height);
+        } else if (rotation == 270) {
+          mapping->area.x1 = width - rect.y2;
+          mapping->area.x2 = MIN (mapping->area.x1 + annot_width, width);
+          mapping->area.y2 = rect.x1;
+          mapping->area.y1 = MAX (0, mapping->area.y2 - annot_height);
+        }
+      } else { /* !flag_no_rotate */
+        if (rotation == 90) {
+          mapping->area.x1 = rect.y1;
+          mapping->area.y1 = height - rect.x2;
+          mapping->area.x2 = mapping->area.x1 + annot_height;
+          mapping->area.y2 = mapping->area.y1 + annot_width;
+        } else if (rotation == 180) {
+          mapping->area.x1 = width - rect.x2;
+          mapping->area.y1 = height - rect.y2;
+          mapping->area.x2 = mapping->area.x1 + annot_width;
+          mapping->area.y2 = mapping->area.y1 + annot_height;
+        } else if (rotation == 270) {
+          mapping->area.x1 = width - rect.y2;
+          mapping->area.y1 = rect.x1;
+          mapping->area.x2 = mapping->area.x1 + annot_height;
+          mapping->area.y2 = mapping->area.y1 + annot_width;
+        }
       }
+    }
 
     map_list = g_list_prepend (map_list, mapping);
   }
@@ -1479,8 +1524,35 @@ poppler_page_free_annot_mapping (GList *list)
   if (G_UNLIKELY (!list))
     return;
 
-  g_list_foreach (list, (GFunc)poppler_annot_mapping_free, NULL);
-  g_list_free (list);
+  g_list_free_full (list, (GDestroyNotify)poppler_annot_mapping_free);
+}
+
+/* Adds or removes (according to @add parameter) the passed in @crop_box from the
+ * passed in @quads and returns it as a new #AnnotQuadrilaterals object */
+static AnnotQuadrilaterals *
+new_quads_from_offset_cropbox (const PDFRectangle* crop_box,
+                               AnnotQuadrilaterals *quads,
+                               gboolean add)
+{
+  int len = quads->getQuadrilateralsLength();
+  auto quads_array = std::make_unique<AnnotQuadrilaterals::AnnotQuadrilateral[]>(len);
+  for (int i = 0; i < len; i++) {
+     if (add) {
+       quads_array[i] = AnnotQuadrilaterals::AnnotQuadrilateral (
+         quads->getX1(i) + crop_box->x1, quads->getY1(i) + crop_box->y1,
+         quads->getX2(i) + crop_box->x1, quads->getY2(i) + crop_box->y1,
+         quads->getX3(i) + crop_box->x1, quads->getY3(i) + crop_box->y1,
+         quads->getX4(i) + crop_box->x1, quads->getY4(i) + crop_box->y1);
+     } else {
+       quads_array[i] = AnnotQuadrilaterals::AnnotQuadrilateral (
+         quads->getX1(i) - crop_box->x1, quads->getY1(i) - crop_box->y1,
+         quads->getX2(i) - crop_box->x1, quads->getY2(i) - crop_box->y1,
+         quads->getX3(i) - crop_box->x1, quads->getY3(i) - crop_box->y1,
+         quads->getX4(i) - crop_box->x1, quads->getY4(i) - crop_box->y1);
+     }
+  }
+
+  return new AnnotQuadrilaterals (std::move(quads_array), len);
 }
 
 /**
@@ -1496,9 +1568,36 @@ void
 poppler_page_add_annot (PopplerPage  *page,
 			PopplerAnnot *annot)
 {
+  double x1, y1, x2, y2;
+  const PDFRectangle *crop_box;
+  const PDFRectangle *page_crop_box;
+
   g_return_if_fail (POPPLER_IS_PAGE (page));
   g_return_if_fail (POPPLER_IS_ANNOT (annot));
 
+  /* Add the page's cropBox to the coordinates of rect field of annot */
+  page_crop_box = page->page->getCropBox ();
+  annot->annot->getRect(&x1, &y1, &x2, &y2);
+  annot->annot->setRect(x1 + page_crop_box->x1,
+                        y1 + page_crop_box->y1,
+                        x2 + page_crop_box->x1,
+                        y2 + page_crop_box->y1);
+
+  AnnotTextMarkup *annot_markup = dynamic_cast<AnnotTextMarkup*>(annot->annot);
+  if (annot_markup) {
+    AnnotQuadrilaterals *quads;
+    crop_box = _poppler_annot_get_cropbox (annot);
+    if (crop_box) {
+      /* Handle hypothetical case of annot being added is already existing on a prior page, so
+       * first remove cropbox of the prior page before adding cropbox of the new page later */
+      quads = new_quads_from_offset_cropbox (crop_box, annot_markup->getQuadrilaterals(), FALSE);
+      annot_markup->setQuadrilaterals( quads );
+    }
+    /* Add to annot's quadrilaterals the offset for the cropbox of the new page */
+    quads = new_quads_from_offset_cropbox (page_crop_box, annot_markup->getQuadrilaterals(), TRUE);
+    annot_markup->setQuadrilaterals( quads );
+  }
+
   page->page->addAnnot (annot->annot);
 }
 
@@ -1551,7 +1650,7 @@ poppler_rectangle_new (void)
 PopplerRectangle *
 poppler_rectangle_copy (PopplerRectangle *rectangle)
 {
-  g_return_val_if_fail (rectangle != NULL, NULL);
+  g_return_val_if_fail (rectangle != nullptr, NULL);
 
   return g_slice_dup (PopplerRectangle, rectangle);
 }
@@ -1603,7 +1702,7 @@ poppler_point_new (void)
 PopplerPoint *
 poppler_point_copy (PopplerPoint *point)
 {
-  g_return_val_if_fail (point != NULL, NULL);
+  g_return_val_if_fail (point != nullptr, NULL);
 
   return g_slice_dup (PopplerPoint, point);
 }
@@ -1656,7 +1755,7 @@ poppler_quadrilateral_new (void)
 PopplerQuadrilateral *
 poppler_quadrilateral_copy (PopplerQuadrilateral *quad)
 {
-  g_return_val_if_fail (quad != NULL, NULL);
+  g_return_val_if_fail (quad != nullptr, NULL);
 
   return g_slice_dup (PopplerQuadrilateral, quad);
 }
@@ -1697,9 +1796,9 @@ poppler_text_attributes_new (void)
 }
 
 static gchar *
-get_font_name_from_word (TextWord *word, gint word_i)
+get_font_name_from_word (const TextWord *word, gint word_i)
 {
-  GooString *font_name = word->getFontName(word_i);
+  const GooString *font_name = word->getFontName(word_i);
   const gchar *name;
   gboolean subset;
   gint i;
@@ -1714,7 +1813,7 @@ get_font_name_from_word (TextWord *word, gint word_i)
     }
   }
   subset = i > 0 && i < font_name->getLength () && font_name->getChar (i) == '+';
-  name = font_name->getCString ();
+  name = font_name->c_str ();
   if (subset)
     name += i + 1;
 
@@ -1725,7 +1824,7 @@ get_font_name_from_word (TextWord *word, gint word_i)
  * Allocates a new PopplerTextAttributes with word attributes
  */
 static PopplerTextAttributes *
-poppler_text_attributes_new_from_word (TextWord *word, gint i)
+poppler_text_attributes_new_from_word (const TextWord *word, gint i)
 {
   PopplerTextAttributes *attrs = poppler_text_attributes_new ();
   gdouble r, g, b;
@@ -2103,7 +2202,7 @@ poppler_annot_mapping_free (PopplerAnnotMapping *mapping)
 void
 poppler_page_get_crop_box (PopplerPage *page, PopplerRectangle *rect)
 {
-  PDFRectangle* cropBox = page->page->getCropBox ();
+  const PDFRectangle* cropBox = page->page->getCropBox ();
   
   rect->x1 = cropBox->x1;
   rect->x2 = cropBox->x2;
@@ -2111,6 +2210,53 @@ poppler_page_get_crop_box (PopplerPage *page, PopplerRectangle *rect)
   rect->y2 = cropBox->y2;
 }
 
+/*
+ * poppler_page_get_bounding_box:
+ * @page: A #PopplerPage
+ * @rect: (out) return the bounding box of the page
+ *
+ * Returns the bounding box of the page, a rectangle enclosing all text, vector
+ * graphics (lines, rectangles and curves) and raster images in the page.
+ * Includes invisible text but not (yet) annotations like highlights and form
+ * elements.
+ *
+ * Return value: %TRUE if the page contains graphics, %FALSE otherwise
+ *
+ * Since: 0.88
+ */
+gboolean
+poppler_page_get_bounding_box (PopplerPage *page,
+                               PopplerRectangle *rect) {
+  Gfx *gfx;
+  BBoxOutputDev *bb_out;
+  bool hasGraphics;
+
+  g_return_val_if_fail(POPPLER_IS_PAGE (page), false);
+  g_return_val_if_fail(rect != nullptr, false);
+
+  bb_out = new BBoxOutputDev(page->page->getCropBox());
+
+  gfx = page->page->createGfx(bb_out,
+                              72.0, 72.0, 0,
+                              false, /* useMediaBox */
+                              true, /* Crop */
+                              -1, -1, -1, -1,
+                              false, /* printing */
+                              nullptr, nullptr);
+  page->page->display(gfx);
+  hasGraphics = bb_out->getHasGraphics();
+  if (hasGraphics) {
+    rect->x1 = bb_out->getX1();
+    rect->y1 = bb_out->getY1();
+    rect->x2 = bb_out->getX2();
+    rect->y2 = bb_out->getY2();
+  }
+
+  delete gfx;
+  delete bb_out;
+  return hasGraphics;
+}
+
 /**
  * poppler_page_get_text_layout:
  * @page: A #PopplerPage
@@ -2169,16 +2315,15 @@ poppler_page_get_text_layout_for_area (PopplerPage       *page,
   TextPage *text;
   PopplerRectangle *rect;
   PDFRectangle selection;
-  int i, j, k;
+  int i, k;
   guint offset = 0;
   guint n_rects = 0;
   gdouble x1, y1, x2, y2;
   gdouble x3, y3, x4, y4;
-  GooList **word_list;
   int n_lines;
 
   g_return_val_if_fail (POPPLER_IS_PAGE (page), FALSE);
-  g_return_val_if_fail (area != NULL, FALSE);
+  g_return_val_if_fail (area != nullptr, FALSE);
 
   *n_rectangles = 0;
 
@@ -2188,18 +2333,17 @@ poppler_page_get_text_layout_for_area (PopplerPage       *page,
   selection.y2 = area->y2;
 
   text = poppler_page_get_text_page (page);
-  word_list = text->getSelectionWords (&selection, selectionStyleGlyph, &n_lines);
+  std::vector<TextWordSelection*>** word_list = text->getSelectionWords (&selection, selectionStyleGlyph, &n_lines);
   if (!word_list)
           return FALSE;
 
   n_rects += n_lines - 1;
   for (i = 0; i < n_lines; i++)
     {
-      GooList *line_words = word_list[i];
-      n_rects += line_words->getLength() - 1;
-      for (j = 0; j < line_words->getLength(); j++)
+      std::vector<TextWordSelection*> *line_words = word_list[i];
+      n_rects += line_words->size() - 1;
+      for (const TextWordSelection *word_sel : *line_words)
         {
-          TextWordSelection *word_sel = (TextWordSelection *)line_words->get(j);
           n_rects += word_sel->getEnd() - word_sel->getBegin();
         }
     }
@@ -2209,11 +2353,11 @@ poppler_page_get_text_layout_for_area (PopplerPage       *page,
 
   for (i = 0; i < n_lines; i++)
     {
-      GooList *line_words = word_list[i];
-      for (j = 0; j < line_words->getLength(); j++)
+      std::vector<TextWordSelection*> *line_words = word_list[i];
+      for (std::size_t j = 0; j < line_words->size(); j++)
         {
-          TextWordSelection *word_sel = (TextWordSelection *)line_words->get(j);
-          TextWord *word = word_sel->getWord();
+          TextWordSelection *word_sel = (*line_words)[j];
+          const TextWord *word = word_sel->getWord();
           int end = word_sel->getEnd();
 
           for (k = word_sel->getBegin(); k < end; k++)
@@ -2230,11 +2374,11 @@ poppler_page_get_text_layout_for_area (PopplerPage       *page,
           rect = *rectangles + offset;
           word->getBBox (&x1, &y1, &x2, &y2);
 
-          if (j < line_words->getLength() - 1)
+          if (j < line_words->size() - 1)
             {
-              TextWordSelection *word_sel = (TextWordSelection *)line_words->get(j + 1);
+              TextWordSelection *next_word_sel = (*line_words)[j + 1];
 
-              word_sel->getWord()->getBBox(&x3, &y3, &x4, &y4);
+              next_word_sel->getWord()->getBBox(&x3, &y3, &x4, &y4);
 	      // space is from one word to other and with the same height as
 	      // first word.
 	      rect->x1 = x2;
@@ -2278,15 +2422,14 @@ poppler_page_get_text_layout_for_area (PopplerPage       *page,
 void
 poppler_page_free_text_attributes (GList *list)
 {
-  if (G_UNLIKELY (list == NULL))
+  if (G_UNLIKELY (list == nullptr))
     return;
 
-  g_list_foreach (list, (GFunc)poppler_text_attributes_free, NULL);
-  g_list_free (list);
+  g_list_free_full (list, (GDestroyNotify)poppler_text_attributes_free);
 }
 
 static gboolean
-word_text_attributes_equal (TextWord *a, gint ai, TextWord *b, gint bi)
+word_text_attributes_equal (const TextWord *a, gint ai, const TextWord *b, gint bi)
 {
   double ar, ag, ab, br, bg, bb;
 
@@ -2355,17 +2498,16 @@ poppler_page_get_text_attributes_for_area (PopplerPage      *page,
 {
   TextPage *text;
   PDFRectangle selection;
-  GooList **word_list;
   int n_lines;
-  PopplerTextAttributes *attrs = NULL;
-  TextWord *word, *prev_word = NULL;
+  PopplerTextAttributes *attrs = nullptr;
+  const TextWord *word, *prev_word = nullptr;
   gint word_i, prev_word_i;
-  gint i, j;
+  gint i;
   gint offset = 0;
-  GList *attributes = NULL;
+  GList *attributes = nullptr;
 
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
-  g_return_val_if_fail (area != NULL, FALSE);
+  g_return_val_if_fail (area != nullptr, nullptr);
 
   selection.x1 = area->x1;
   selection.y1 = area->y1;
@@ -2373,16 +2515,16 @@ poppler_page_get_text_attributes_for_area (PopplerPage      *page,
   selection.y2 = area->y2;
 
   text = poppler_page_get_text_page (page);
-  word_list = text->getSelectionWords (&selection, selectionStyleGlyph, &n_lines);
+  std::vector<TextWordSelection*>** word_list = text->getSelectionWords (&selection, selectionStyleGlyph, &n_lines);
   if (!word_list)
-          return NULL;
+          return nullptr;
 
   for (i = 0; i < n_lines; i++)
     {
-      GooList *line_words = word_list[i];
-      for (j = 0; j < line_words->getLength(); j++)
+      std::vector<TextWordSelection*> *line_words = word_list[i];
+      for (std::size_t j = 0; j < line_words->size(); j++)
         {
-          TextWordSelection *word_sel = (TextWordSelection *)line_words->get(j);
+          TextWordSelection *word_sel = (*line_words)[j];
           int end = word_sel->getEnd();
 
           word = word_sel->getWord();
@@ -2401,7 +2543,7 @@ poppler_page_get_text_attributes_for_area (PopplerPage      *page,
               prev_word_i = word_i;
             }
 
-          if (j < line_words->getLength() - 1)
+          if (j < line_words->size() - 1)
             {
               attrs->end_index = offset;
               offset++;
